#     Copyright 2023, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Utils module to provide helper for our common json operations.

"""

from __future__ import absolute_import

import json

from .FileOperations import getFileContents, openTextFile


def loadJsonFromFilename(filename):
    try:
        return json.loads(getFileContents(filename))
    except ValueError:
        return None


def writeJsonToFilename(filename, contents):
    with openTextFile(filename, "w") as output:
        json.dump(contents, output)
